# -*- coding: utf-8 -*-

# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Web client based on pycurl, with an API similar to urllib2."""

import httplib

from email import parser
from StringIO import StringIO

import pycurl


class HTTPError(Exception):
    """Error that happens while doing a web request."""

    def __init__(self, url, code, message, headers=None, fp=None):
        """Initialize this instance."""
        super(HTTPError, self).__init__()
        self.code = code
        self.message = message
        if fp:
            self.fp = fp
        else:
            self.fp = StringIO(message)

    def read(self):
        """Return the error message."""
        return self.fp.read()

    def __str__(self):
        return "<%s: %s>" % (self.code, self.message)


class UnauthorizedError(HTTPError):
    """An HTTP error when there's an auth problem."""


class Request(object):
    """An HTTP request object."""

    def __init__(self, url, data=None, headers=None):
        """Initialize this instance."""
        super(Request, self).__init__()
        self.url = url
        self.data = data
        self.headers = headers

    def get_full_url(self):
        """Return the url."""
        return self.url

    def get_data(self):
        """Return the data."""
        return self.data


class Response(StringIO):
    """An HTTP response object."""

    code = -1
    headers = None

    def __init__(self, url):
        """Initialize this instance."""
        StringIO.__init__(self)
        self.url = url

    def finish(self, code, headers):
        """Finish a response and rewind it."""
        self.code = code
        self.headers = headers
        self.seek(0)


class HeaderParser(parser.FeedParser):
    """A parser for the HTTP headers, based on the stdlib email package."""

    first_line = True

    def feed(self, data):
        """Feed some data, but chomp the first line."""
        if self.first_line:
            self.first_line = False
            return
        parser.FeedParser.feed(self, data)


def urlopen(request, data=None):
    """Open a given url using curl."""
    if isinstance(request, basestring):
        request = Request(request, data)

    curl = pycurl.Curl()
    try:
        request_headers = []
        if isinstance(request.url, unicode):
            request.url = request.url.encode("utf-8")
        curl.setopt(pycurl.URL, request.url)
        if request.headers:
            for key, value in request.headers.items():
                request_headers.append("%s: %s" % (key, value))
        curl.setopt(pycurl.HTTPHEADER, request_headers)
        response = Response(request.url)
        response_headers_parser = HeaderParser()
        curl.setopt(pycurl.WRITEFUNCTION, response.write)
        curl.setopt(pycurl.HEADERFUNCTION, response_headers_parser.feed)
        curl.setopt(pycurl.FOLLOWLOCATION, 1)
        curl.setopt(pycurl.MAXREDIRS, 5)
        curl.setopt(pycurl.SSL_VERIFYPEER, 1)
        curl.setopt(pycurl.SSL_VERIFYHOST, 2)
        if request.data:
            curl.setopt(pycurl.POST, 1)
            curl.setopt(pycurl.POSTFIELDS, request.data)
        curl.perform()
        code = curl.getinfo(pycurl.HTTP_CODE)
        response_headers = response_headers_parser.close()
        response.finish(code, response_headers)
    except pycurl.error as e:
        raise HTTPError(request.url, e[0], curl.errstr())
    else:
        if code in (200, 0):
            return response
        else:
            if code == 401:
                errorclass = UnauthorizedError
            else:
                errorclass = HTTPError
            message = httplib.responses.get(code, "Unknown error")
            raise errorclass(request.url, code, message, response)
    finally:
        curl.close()
